/* -----------------------------------------

   Create deltaC x-tables for differential heavy-quark pair production

   ----------------------------------------- */
#include <iomanip>
#include <iostream>
#include <fstream>
#include <string>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <sys/time.h>
#include <getopt.h>

#include "hell.hh"
#include "QQbarPair.hh"


using namespace std;


double cubicinterpolate(double x, int N, double* vx, double *y) {
  if(N<4) {
    cout << "Error: vector too short for cubic interpolation" << endl;
    exit(57);
  }
  if(x<vx[0]) {
    cout << "Error: interpolation outside interpolation range: x="<<x<<" < xmax="<<vx[0] << endl;
    //exit(58);
    x=vx[0];
  }
  if(x>vx[N-1]) {
    cout << "Error: interpolation outside interpolation range: x="<<x<<" > xmax="<<vx[N-1] << endl;
    //exit(59);
    x=vx[N-1];
  }
  int k = 1;
  for(int i=1; i<N-2; i++) {
    if(vx[i] > x) break;
    k = i;
  }
  //cout << k << "  " << x << "  " << vx[k] << "  " << vx[k+1] << endl;
  //if(k>N-3) k=N-3;
  double v[4] = { vx[k-1], vx[k], vx[k+1], vx[k+2] };
  //return (x-v[2])*y[k]/(v[1]-v[2]) + (x-v[1])*y[k+1]/(v[2]-v[1]);
  double res = (x-v[1])*(x-v[2])*(x-v[3])*y[k-1]/(v[0]-v[1])/(v[0]-v[2])/(v[0]-v[3])
    +          (x-v[0])*(x-v[2])*(x-v[3])*y[k  ]/(v[1]-v[0])/(v[1]-v[2])/(v[1]-v[3])
    +          (x-v[0])*(x-v[1])*(x-v[3])*y[k+1]/(v[2]-v[0])/(v[2]-v[1])/(v[2]-v[3])
    +          (x-v[0])*(x-v[1])*(x-v[2])*y[k+2]/(v[3]-v[0])/(v[3]-v[1])/(v[3]-v[2]);
  if(res!=res) {
    cout << "cubicinterpolate is NaN" << endl;
  }
  return res;
}


string sas(double as) {
  ostringstream os;
  int ias = int(1000*as+0.1);
  if(1000*as-ias>0.5) ias += 1;
  if     (ias<10  ) os << "000" << ias;
  else if(ias<100 ) os << "00"  << ias;
  else if(ias<1000) os << "0"   << ias;
  else os << ias;
  return os.str();
}
string sx(double x) {
  ostringstream os;
  int ix = int(1e8*x+0.1);
  if(1e8*x-ix>0.5) ix += 1;
  if     (ix<1e1) os << "00000000"  << ix;
  else if(ix<1e2) os << "0000000"  << ix;
  else if(ix<1e3) os << "000000"  << ix;
  else if(ix<1e4) os << "00000"  << ix;
  else if(ix<1e5) os << "0000"  << ix;
  else if(ix<1e6) os << "000"  << ix;
  else if(ix<1e7) os << "00"  << ix;
  else if(ix<1e8) os << "0"   << ix;
  else os << ix;
  return os.str();
}


void print_usage() {
  cout << endl
       << "Usage:" << endl
       << " -a  --alphas <as>      set alpha_s" << endl
       << " -m  --invM <Q>         set the pair invariant mass" << endl
       << " -F  --muF <muF/Q>      set muF/Q" << endl
       << " -R  --RCmode           use the RC variant r->as*b0 (for estimating the uncertainty)" << endl
       << " -v  --variation        use the g+ variant \"Airy\" (for estimating the uncertainty)" << endl
       << " -r  --redo             do NOT recompute the resummed part with 2 M's" << endl
       << " -L  --LLp              use LLp (default is NLL)" << endl
       << endl;
  exit(0);
  return;
}

void read_arguments(int argc, char* argv[], double &as, double &mH, double &muFrat, int &mode, bool &redo, bool &useLLp, double &x, double &pt) {
  const char* const short_options = "ha:m:F:RvrLx:p:";
  const struct option long_options[] = { { "help", 0, NULL, 'h' },
					 { "alphas", 1, NULL, 'a' },
					 { "mH", 1, NULL, 'm' },
					 { "muF", 1, NULL, 'F' },
					 { "RCmode", 0, NULL, 'R' },
					 { "variation", 0, NULL, 'v' },
					 { "redo", 0, NULL, 'r' },
					 { "LLp", 0, NULL, 'L' },
					 { "xx", 1, NULL, 'x' },
					 { "pt", 1, NULL, 'p' },
					 { NULL, 0, NULL, 0 } };
  int next_option;
  ostringstream sset;
  do {
    next_option = getopt_long (argc, argv, short_options, long_options, NULL);
    switch (next_option) {
    case 'h':
      print_usage();
    case 'a':
      as = strtod(optarg, NULL);
      break;
    case 'p':
      pt = strtod(optarg, NULL);
      break;
    case 'x':
      x = strtod(optarg, NULL);
      break;
    case 'm':
      mH = strtod(optarg, NULL);
      break;
    case 'F':
      muFrat = strtod(optarg, NULL);
      break;
    case 'v':
      mode = 1;
      break;
    case 'R':
      mode = 2;
      break;
    case 'r':
      redo = false;
      break;
    case 'L':
      useLLp = true;
      break;
    case '?':
      print_usage();
    case -1: break;
    default: abort();
    }
  }
  while (next_option != -1);
  return;
}



int main (int argc, char* argv[]) {

  struct timeval t0, t1;
  gettimeofday(&t0,NULL);

  double m  = 4.6;
  //double mH = 20;
  double mH = m;
  double x    = 1e-5;
  double rapY = 4;
  double pt   = 2;
  double as0 = 0.22; //0.218323; //0.21841 //0.15;
  int nf = 4;
  int localRCmode = 0;
  double muFrat = 1;
  bool forceredo = true;
  bool useLLp = false;
  //
  read_arguments(argc, argv, as0, mH, muFrat, localRCmode, forceredo, useLLp, x, pt);

  double rapYmax = 0;
  if(pt>=0) {
    //if(x*(1+pt*pt/mH/mH) > 0.25) {  // old condition, where did it come from?
    //  cout << "Error: x*(1+pthat2) = " << x*(1+pt*pt/mH/mH) << " must be smaller than 1/4" << endl;
    //  exit(1);
    //}
    if(pt > 1/x-2/sqrt(x)) {
      cout << "Error: pt = " << pt << " must be smaller than 1/x-2/sqrt(x)=" << 1/x-2/sqrt(x) << endl;
      exit(1);
    }
    double pthat2 = pt*pt/mH/mH;
    double A2 = x*(1+pthat2);
    //rapYmax = log(1./2./sqrt(A2)+sqrt(1./4./A2-1));
    rapYmax = log((1+x*pthat2)/2./sqrt(A2)+sqrt(pow(1+x*pthat2,2)/4./A2-1));
    //rapYmax *= 0.9999;
  } else {
    rapYmax = -log(x)/2.;
  }
  cout << "Ymax = " << rapYmax << endl;

  HELL::SetLLpMode(useLLp);
  HELL::LogOrder ord = HELL::NLL;



  // x grid
  const int Np = 370;
  double x_min = 1e-9;
  double x_mid = 0.1;
  double x_max = 0.99+0.01;

  cout << "nf = " << nf << endl;
  HELL::HELLnf sxD(nf, ord, "./data/");
  sxD.SetLargexDamping();
  sxD.SetMomentumConservation(); // has no effect
  sxD.SetEvaluationMethod(1);
  sxD.SetFactorizationScale(muFrat);
  if(localRCmode) sxD.SetRCmode(1);
  vector<double> as1;
  sxD.GetAvailableAlphas(as1);
  //
  for(unsigned int a=0; a<as1.size(); a++) {
    double as = as1[a];
    if(as0>0 && fabs(as-as0)>1e-6) continue;
    if(as0>0) cout << "as = " << as << endl;
    as = int(1000*as+0.1)/1000.;
    //cout << setprecision(10) << 1000*as << "  " << sas(as) << endl;
    ostringstream filename;
    filename << "./output_for_HELLx/NLL" << (useLLp ? "-LLp" : "") << "/xtableCQQbarSQ";
    if(localRCmode) filename << "var" << localRCmode;
    filename << "_muF" << 1000*muFrat << "_NLL_nf" << nf << "_alphas" << sas(as) << "_pt" << pt << "_x" << sx(x) << ".table";
    //
    ofstream ofile(filename.str().c_str());
    ofile << VERSIONx << endl;
    ofile << "# "
      << setw(15) << Np
      << setw(15) << x
      << setw(15) << rapYmax
      << endl;
    /*
    ofile << setw(15) << c10
	  << setw(15) << c20
	  << setw(15) << c11
	  << setw(15) << c30
	  << setw(15) << c21
	  << endl;
    */
    //
    const int Npreg = 21;
    double dCr[Npreg], rapYreg[Npreg];
    for(int i=0; i<Npreg; i++) {
      rapY = i*rapYmax/(Npreg-1.);
      rapYreg[i] = rapY;
      dCr[i] = sxD.deltaQQbarSingleQuark(as, x, m, rapY, pt);
      cout << "-";
      cout.flush();
    }
    cout << endl;
    
    double dCreg[2*Np+1], dCaux[2*Np+1], dCauxOrdAs[2*Np+1], rapYvec[2*Np+1];
    //cout << "ciao" << endl;
    for(int i=0; i<=Np; i++) {
      rapY = i*rapYmax/Np;
      rapYvec[Np+i] =  rapY;
      rapYvec[Np-i] = -rapY;
      //
      dCreg        [Np+i] = cubicinterpolate(rapY, Npreg, rapYreg, dCr); //sxD.deltaQQbarSingleQuark(as, x, m, rapY, pt);
      if(i>0) dCreg[Np-i] = dCreg[Np+i];
      dCaux        [Np+i] = sxD.deltaQQbarSingleQuarkAux(as, x, m,  rapY, pt);
      if(i>0) dCaux[Np-i] = sxD.deltaQQbarSingleQuarkAux(as, x, m, -rapY, pt);
      dCauxOrdAs        [Np+i] = sxD.deltaQQbarSingleQuarkAuxOrdAs(as, x, m,  rapY, pt);
      if(i>0) dCauxOrdAs[Np-i] = sxD.deltaQQbarSingleQuarkAuxOrdAs(as, x, m, -rapY, pt);
      cout << "-";
      cout.flush();
    }
    //
    for(int i=0; i<2*Np+1; i++) {
      ofile
	<< setw(15) << rapYvec[i]
	<< setw(15) << dCreg[i]
	<< setw(15) << dCaux[i]
	<< setw(15) << dCaux[2*Np-i]
	<< setw(15) << dCauxOrdAs[i]
	<< setw(15) << dCauxOrdAs[2*Np-i]
	<< endl;
    }
    //
    ofile.close();
  }
  cout << endl;


  // Finish time
  gettimeofday(&t1,NULL);
  double t=t1.tv_sec-t0.tv_sec+(t1.tv_usec-t0.tv_usec)*0.000001;
  cout << "Total time: " << t << "s" << endl;

  return 0;

}
