#pragma once
#include <vector>
#include <iostream>
#include <fstream>
#include <gsl/gsl_spline.h>
#include "types.hh"

namespace HELL {

  //typedef Complex (*MellinFunction)(Complex,Complex,Complex,Real,Real,Complex,void*);
  typedef Complex (*MellinFunction)(Complex,Complex,Complex,Real,Real,void*);


  class GammaHellTable {
  public:
    GammaHellTable() : var_(false) {}
    GammaHellTable & read(std::istream & in);
    unsigned getTableSize() const { return gamma_.size(); }
    Complex getN(unsigned i) const;
    Complex getElement(unsigned i) const { return (var_ ? gammavar_[i] : gamma_[i]); }
    void setVariation(bool var=true) { var_ = var; }
    std::string getVersion() const { return version_; }
  private:
    struct Pole {
      Real pole;
      Real residue;
    };
    typedef std::vector<Pole>::const_iterator Poles;
    friend class HellTableP;
    Complex eith_;  // exp(i*theta), i.e. the slope of the integration path
    std::vector<Pole> poles_;
    std::vector<Complex> gamma_, gammavar_;
    Real alphas_, nf_, step_, umax_, N0_;
    bool var_;
    std::string version_;
  };




  /*
    At the moment this class must be used in the following order:
    1a) readTable(...)
    1b) readDotTable(...)   this is optional, if not done sets gammadot=0
    2)  setMellinFunction(...)    now mandatory
    3)  eval(...)
  */
  class HellTableP {
  public:
    //HellTableP() : g_(), g_dot_(), g_nlo_(), err_(0), use_err_(false) { reMFspline_ = new gsl_spline; imMFspline_ = new gsl_spline; }
    HellTableP() : g_(), g_dot_(), err_(0), use_err_(false) { reMFspline_ = new gsl_spline; imMFspline_ = new gsl_spline; }
    ~HellTableP() { delete reMFspline_; delete imMFspline_; }

    HellTableP & readTable   (std::string in);
    HellTableP & readTable   (std::istream & in) { g_    .read(in); return *this; }
    HellTableP & readDotTable(std::string in);
    HellTableP & readDotTable(std::istream & in) { g_dot_.read(in); return *this; }
    //HellTableP & readNLOTable(std::string in);
    //HellTableP & readNLOTable(std::istream & in) { g_nlo_.read(in); return *this; }

    HellTableP & setErrorEstimation(bool val) { use_err_ = val; return *this; }
    HellTableP & setMellinFunction(MellinFunction mf, void *pars=NULL);
    //HellTableP & setVariation(bool var=true) { g_.setVariation(var); g_dot_.setVariation(var); g_nlo_.setVariation(var); return *this; }
    HellTableP & setVariation(bool var=true) { g_.setVariation(var); g_dot_.setVariation(var); return *this; }

    Real getError() const { return err_/PI; }
    Real getAlphaS() const { return g_.alphas_; }
    Real getNF() const { return g_.nf_; }

    Real operator()(Real z) { return eval(z); }
    Real eval(Real z);

    unsigned getTableSize() const;
    // get value of N corresponding to element i, for i in the interval [0, getTableSize()-1]
    Complex getN(unsigned i) const { return g_.getN(i); }
    Complex getGamma(unsigned i) const { return g_.getElement(i) + getNPoles_(g_,getN(i)); }
    Complex getDotGamma(unsigned i) const { return g_dot_.getTableSize() ? g_dot_.getElement(i) + getNPoles_(g_dot_,getN(i)) : 0; }
    //Complex getNLOGamma(unsigned i) const { return g_nlo_.getTableSize() ? g_nlo_.getElement(i) + getNPoles_(g_nlo_,getN(i)) : 0; }
    Complex getMellinFunction(int i) {return MFvals_[i];}
    Complex getMellinFunctionSpline(double u);

    Real getIntegrand(Real z, int i);
    Real getIntegrand(MellinFunction cf, Real z, int i, void *pars=NULL);
    Real GaussIntegrand(Real t);
    void setx(Real x) { x_ = x; }
    Real getBatemanPole() { return g_.poles_[0].pole; }

  protected:
    Real getJac(unsigned i) const;
    Real evalGauss(Real z, int &status);
    Real evalSimpson(Real z);
    Real x_;
    friend double integrand(double x, void *p);

    Complex getNPoles_(GammaHellTable g, Complex nt) const;
   
  private:
    typedef GammaHellTable::Poles Poles;
    GammaHellTable g_, g_dot_;//, g_nlo_;
    Real err_;
    bool use_err_;
    std::vector<Complex> MFvals_;
    gsl_spline *reMFspline_, *imMFspline_;

  };
  
  
} // namespace HELL

